# Adapted from
# https://github.com/lm-sys/FastChat/blob/168ccc29d3f7edc50823016105c024fe2282732a/fastchat/protocol/openai_api_protocol.py

# Copyright (c) 2021 - present / Neuralmagic, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import uuid
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field


def random_uuid() -> str:
    return str(uuid.uuid4().hex)


class LogProbs(BaseModel):
    text_offset: List[Optional[int]] = Field(default_factory=list)
    token_logprobs: List[float] = Field(default_factory=list)
    tokens: List[str] = Field(default_factory=list)


class ErrorResponse(BaseModel):
    """
    Structure of the error messages returned for any errors returned as part of the
    OpenAI integration.
    """

    object: str = "error"
    message: str
    type: str
    param: Optional[str] = None
    code: Optional[str] = None


class ModelPermission(BaseModel):
    """
    Expected structure of the permission for each model (included as part of the
    ModelCard) supported for inference.
    """

    id: str = Field(default_factory=lambda: f"modelperm-{random_uuid()}")
    object: str = "model_permission"
    created: int = Field(default_factory=lambda: int(time.time()))
    allow_create_engine: bool = False
    allow_sampling: bool = True
    allow_logprobs: bool = True
    allow_search_indices: bool = False
    allow_view: bool = True
    allow_fine_tuning: bool = False
    organization: str = "*"
    group: Optional[str] = None
    is_blocking: str = False


class ModelCard(BaseModel):
    """
    Expected structure for each item in the ModelList returned from the /v1/models
    endpoint, as part of the OpenAI integration. Each model available for inference
    will have its own ModelCard.
    """

    id: str
    object: str = "model"
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: str = "neuralmagic"
    root: Optional[str] = None
    parent: Optional[str] = None
    permission: List[ModelPermission] = Field(default_factory=list)


class ModelList(BaseModel):
    """
    Expected structure for the response from the /v1/models endpoint,
    as part of the OpenAI integration. Includes a list structure, containing a ModelCard
    for each model available to run inference.
    """

    object: str = "list"
    data: List[ModelCard] = Field(default_factory=list)


class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0
    completion_tokens: Optional[int] = 0


class ChatCompletionRequest(BaseModel):
    """
    Expected structure for requests sent to the /v1/chat/completions endpoint,
    as part of the OpenAI integration. This payload is mapped to the input
    expected by the TextGenerationPipeline.
    """

    model: Optional[str] = None
    messages: Union[str, List[Dict[str, str]], Dict[str, str]]
    temperature: Optional[float] = 0.7
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    max_tokens: Optional[int] = 16
    stop: Optional[Union[str, List[str]]] = Field(default_factory=list)
    stream: Optional[bool] = False
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None
    # Additional parameters
    top_k: Optional[int] = -1
    ignore_eos: Optional[bool] = False
    add_generation_prompt: Optional[bool] = True


class CompletionRequest(BaseModel):
    """
    Expected structure for requests sent to the /v1/completions endpoint,
    as part of the OpenAI integration. This payload is mapped to the input
    expected by the TextGenerationPipeline.
    """

    model: str
    prompt: Union[List[int], List[List[int]], str, List[str]]
    suffix: Optional[str] = None
    max_tokens: Optional[int] = 16
    temperature: Optional[float] = 1.0
    top_p: Optional[float] = 1.0
    n: Optional[int] = 1
    stream: Optional[bool] = False
    logprobs: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = Field(default_factory=list)
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    logit_bias: Optional[Dict[str, float]] = None
    user: Optional[str] = None
    # Additional parameters
    top_k: Optional[int] = -1
    ignore_eos: Optional[bool] = False


class ChatMessage(BaseModel):
    role: str
    content: str


class ChatCompletionResponseChoice(BaseModel):
    """
    Expected structure of each response choice. This schema defines the structure
    of each item in the `choices` field of the ChatCompletionResponse.
    """

    message: ChatMessage
    finish_reason: Optional[
        Literal["stop", "length", "callback", "capacity", "max_new_tokens"]
    ] = None


class ChatCompletionResponse(BaseModel):
    """
    Expected structure for responses returned from the /v1/chat/completions endpoint,
    as part of the OpenAI integration.
    """

    id: str = Field(default_factory=lambda: f"chatcmpl-{random_uuid()}")
    object: str = "chat.completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo


class CompletionResponseChoice(BaseModel):
    """
    Expected structure of each response choice. This schema defines the structure
    of each item in the `choices` field of the CompletionResponse.
    """

    text: str
    logprobs: Optional[LogProbs] = None
    finish_reason: Optional[
        Literal["stop", "length", "callback", "capacity", "max_new_tokens"]
    ] = None


class CompletionResponse(BaseModel):
    """
    Expected structure for responses returned from the /v1/completions endpoint,
    as part of the OpenAI integration.
    """

    id: str = Field(default_factory=lambda: f"cmpl-{random_uuid()}")
    object: str = "text_completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[CompletionResponseChoice]
    usage: UsageInfo


class CompletionResponseStreamChoice(BaseModel):
    text: str
    logprobs: Optional[LogProbs] = None
    finish_reason: Optional[
        Literal["stop", "length", "callback", "capacity", "max_new_tokens"]
    ] = None


class CompletionStreamResponse(BaseModel):
    """
    Expected structure for responses returned from the /v1/completions endpoint
    when streaming is enabled, as part of the OpenAI integration.
    """

    id: str = Field(default_factory=lambda: f"cmpl-{random_uuid()}")
    object: str = "text_completion"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[CompletionResponseStreamChoice]


class DeltaMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = None


class ChatCompletionResponseStreamChoice(BaseModel):
    delta: DeltaMessage
    finish_reason: Optional[
        Literal["stop", "length", "callback", "capacity", "max_new_tokens"]
    ] = None


class ChatCompletionStreamResponse(BaseModel):
    """
    Expected structure for responses returned from the /v1/chat/completions endpoint
    when streaming is enabled, as part of the OpenAI integration.
    """

    id: str = Field(default_factory=lambda: f"chatcmpl-{random_uuid()}")
    object: str = "chat.completion.chunk"
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseStreamChoice]
